from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Iterable, Optional

import h5py
import numpy as np

from opencosmo.io.schema import FileEntry, Schema, make_schema
from opencosmo.io.verify import ZeroLengthError, verify_file
from opencosmo.io.writer import ColumnCombineStrategy, ColumnWriter
from opencosmo.mpi import MPI, get_comm_world

if TYPE_CHECKING:
    from pathlib import Path

    from opencosmo.io.schema import Schema


"""
When working with MPI, datasets are chunked across ranks. Here we combine the schemas
from several ranks into a single schema that can be allocated by rank 0. Each 
rank will then write its own data to the specific section of the file 
it is responsible for.

When writing data with MPI, there are basically 3 things we have to verify in order to 
determine if everything is valid.

1. Is the top-level file structure the same for all ranks (e.g. lightcone? dataset?).
2. Do all columns that are going to be written to by two or more ranks have the same data type and compatible shapes?
3. Is metadata consistent across ranks? If not, are there rules in place to combine/update the fields?

If all three of these checks pass, it is guaranteed we can create a schema that can accomodate the data being written.

File schemas are simply collections of columns and metadata. Columns contain:

1. A reference to the underying data that will be written (either an h5py dataset or a numpy array)
2. An index which tells us which elements in 1 we are going to actually write
3. Possibly an output index, which tells us where in the output we are going to actually write to.
3. Possibly a function to update those values before writing. For example, a spatial index should be summed across ranks rather than concatenated.

In order to avoid MPI deadlocks, we always sort columns in alphabetical order before performing operations on the file.
"""


class CombineState(Enum):
    VALID = 1
    ZERO_LENGTH = 2
    INVALID = 3


def write_parallel(file: Path, file_schema: Schema):
    """
    Main entry point for writing data in parallel. Proceeds
    in three steps:

    1. Verify the file schemas are valid separately
    2. Recursively verify the schemas can be made consistent across ranks
    3. Recursively write the data.
    """
    comm = get_comm_world()
    if comm is None:
        raise ValueError("Got a null comm!")
    paths = set(comm.allgather(file))
    if len(paths) != 1:
        raise ValueError("Different ranks recieved a different path to output to!")

    try:
        verify_file(file_schema)  # Initial verification
        results = comm.allgather(CombineState.VALID)
    except ValueError:
        results = comm.allgather(CombineState.INVALID)
    except ZeroLengthError:
        results = comm.allgather(CombineState.ZERO_LENGTH)
    if any(rs == CombineState.INVALID for rs in results):
        raise ValueError("One or more ranks recieved invalid schemas!")

    has_data = [i for i, state in enumerate(results) if state == CombineState.VALID]
    if len(has_data) == 0:
        raise ValueError("No ranks have any data to write!")

    group = comm.Get_group()
    new_group = group.Incl(has_data)
    new_comm = comm.Create(new_group)
    if new_comm == MPI.COMM_NULL:
        return cleanup_mpi(comm, new_comm, new_group)

    verify_schemas(file_schema, new_comm)
    offsets = __get_all_offsets(file_schema, new_comm, "")
    if new_comm.Get_rank() == 0:
        with h5py.File(file, "w") as f:
            __allocate(file_schema, f, new_comm)
    else:
        __allocate(file_schema, None, new_comm)

    try:
        with h5py.File(file, "a", driver="mpio", comm=new_comm) as f:
            __write_parallel(file_schema, f, offsets, new_comm)

    except ValueError:  # parallell hdf5 not available
        schema = __replace_writers_with_updates(file_schema, new_comm)
        __write_serial(schema, file, offsets, new_comm)
    cleanup_mpi(comm, new_comm, new_group)


def cleanup_mpi(comm_world: MPI.Comm, comm_write: MPI.Comm, group_write: MPI.Group):
    comm_world.Barrier()
    if comm_write != MPI.COMM_NULL:
        comm_write.Free()
    group_write.Free()


def get_all_keys(data: dict, comm: Optional[MPI.Comm]):
    """
    Return all keys in the dictionary across all ranks, sorted
    alphabetically. When defining the file structure, we have to iterate
    through the schemas in the same order across all ranks, including
    when one rank doesn't have a given child.
    """
    data_names = set(data.keys())
    if comm is None:
        return sorted(list(data_names))

    all_data_names: Iterable[str]
    all_data_names = data_names.union(*comm.allgather(data_names))
    all_data_names = list(all_data_names)
    all_data_names.sort()
    return all_data_names


def verify_schemas(schema: Schema, comm: MPI.Comm) -> None:
    """
    By this stage, we know that all the ranks that are participating have a valid
    file schema. We now need to verify that they can be made consistent across ranks.

    This is done recursively. For each schema, we verify its columns and its attributes.
    Then we proceed to its children. Ranks that do not have data for a given child
    are simply excluded from the verification process.

    """

    if comm.Get_size() == 1:  # this shouldn't happen, but include anyway
        return

    file_types = set(comm.allgather(schema.type))
    if len(file_types) > 1:
        raise ValueError(
            "Unable to combine file schemas, as they do not have the same type!"
        )

    verify_columns(schema.columns, comm)
    verify_attributes(schema.attributes, comm)
    all_child_names = get_all_keys(schema.children if schema is not None else {}, comm)

    for child_name in all_child_names:
        has_child = comm.allgather(child_name in schema.children)
        if all(has_child):
            new_comm = comm
        else:
            ranks_to_include = [i for i in range(len(has_child)) if has_child[i]]
            group = comm.Get_group()
            new_group = group.Incl(ranks_to_include)
            new_comm = comm.Create(new_group)
        if child_name in schema.children:
            verify_schemas(schema.children[child_name], new_comm)


def verify_columns(columns: dict[str, ColumnWriter], comm: MPI.Comm):
    """
    Verify a group of columns in a schema. Note, the single-threaded verification
    process already ensures that all columns in the data group have the same
    length in any given rank, which guarantees this will also be the case when
    we combine columns across ranks. All we have to check here is:

    1. Are the shapes consistent
    2. Are the data types consistent
    3. Are the column attributes consistent

    """
    all_column_names = get_all_keys(columns, comm)
    for colname in all_column_names:
        if colname not in columns:
            # Simply ignore a rank that doesn't have this column
            colmeta = comm.allgather(None)
        else:
            column = columns[colname]
            data_to_send = (
                column.combine_strategy,
                column.shape,
                column.dtype,
                column.attrs,
            )
            colmeta = comm.allgather(data_to_send)

        combine_strategies = set([cm[0] for cm in colmeta if cm is not None])
        if len(combine_strategies) > 1:
            raise ValueError("Combine strategy must be the same accross ranks!")

        shapes = set([cm[1][1:] for cm in colmeta if cm is not None])
        if len(shapes) > 1:
            raise ValueError(
                f"Column {colname} did not have consistent shapes across ranks!"
            )
        dtype_kinds = set([cm[2].kind for cm in colmeta if cm is not None])
        if len(dtype_kinds) > 1:
            # We allow type promotion, though there are very few cases where
            # It would be necessary.
            raise ValueError(
                f"Column {colname} did not have consistent dtypes across ranks!"
            )
        attrs = [cm[3] for cm in colmeta if cm is not None]
        if any(attr_set != attrs[0] for attr_set in attrs[1:]):
            raise ValueError("Metadata was not consistent across ranks!")


def verify_attributes(metadata: dict[str, Any], comm: MPI.Comm):
    all_metadata = comm.allgather(metadata)
    if not all(md == all_metadata[0] for md in all_metadata[1:]):
        raise ValueError("Not all ranks recieved the same metadata!")


def __write_parallel(
    schema: Schema,
    group: h5py.File | h5py.Group,
    offsets: dict,
    comm: Optional[MPI.Comm],
):
    """
    Used with both the parallel and serial version, though the later passes through
    __write_serial.
    """
    __write_columns(schema, group, offsets, comm)
    all_child_names = get_all_keys(schema.children, comm)
    for cn in all_child_names:
        child_schema = schema.children.get(cn, make_schema(cn, FileEntry.EMPTY))
        new_group = group[cn]
        __write_parallel(child_schema, new_group, offsets, comm)


def __write_serial(schema: Schema, file_path: Path, offsets: dict, comm: MPI.Comm):
    """
    We do NOT have parallel hdf5, so we have to write one rank at a time.
    """
    for i in range(comm.Get_size()):
        if i == comm.Get_rank():
            with h5py.File(file_path, "a") as f:
                __write_parallel(schema, f, offsets, None)

        comm.Barrier()


def __replace_writers_with_updates(schema: Schema, comm: MPI.Comm):
    """
    For columns that require updates, compute the update and replace. The most common form
    of updated is "start/size" indexes, since "start" must be updated consistently across
    ranks.

    """
    colnames = get_all_keys(schema.columns, comm)
    for cn in colnames:
        colwriter = schema.columns.get(cn)

        participating = comm.allgather(colwriter is not None)
        if all(participating):
            new_comm = comm
        else:
            participating_ranks = [
                i for i in range(len(participating)) if participating[i]
            ]
            group = comm.Get_group()
            new_group = group.Incl(participating_ranks)
            new_comm = comm.Create(new_group)
        if colwriter is None:
            continue

        has_update = new_comm.allgather(
            colwriter is not None and colwriter.has_transformation
        )
        if any(has_update) and not all(has_update):
            raise ValueError("Update was not consistent across ranks!")
        elif not any(has_update):
            continue
        assert colwriter is not None
        data = colwriter.get_data(new_comm)
        new_writer = ColumnWriter.from_numpy_array(
            data, colwriter.combine_strategy, colwriter.attrs
        )
        schema.columns[cn] = new_writer

    child_names = get_all_keys(schema.children, comm)
    for cn in child_names:
        child_schema = schema.children.get(cn, make_schema(cn, FileEntry.EMPTY))
        new_child_schema = __replace_writers_with_updates(child_schema, comm)
        schema.children[cn] = new_child_schema
    return schema


def __get_all_offsets(schema: Schema, comm: MPI.Comm, name: str):
    """
    Get the rank-wise offset for every column.
    """
    output = {}
    all_column_names = get_all_keys(schema.columns, comm)
    for colname in all_column_names:
        writer = schema.columns.get(colname)
        offset = get_column_offset(writer, comm)
        key = "/".join([name, colname])
        output[key] = offset

    all_child_names = get_all_keys(schema.children, comm)
    for cn in all_child_names:
        key = "/".join([name, cn])
        child_schema = schema.children.get(cn, make_schema(cn, FileEntry.EMPTY))
        child_offsets = __get_all_offsets(child_schema, comm, key)
        output.update(child_offsets)
    return output


def __allocate(schema: Schema, group: Optional[h5py.File | h5py.Group], comm: MPI.Comm):
    """
    Allocate the file.
    """
    all_column_names = get_all_keys(schema.columns, comm)
    for column_name in all_column_names:
        column_writer = schema.columns.get(column_name)
        __allocate_column(column_name, column_writer, group, comm)
    __write_metadata(schema, group, comm)

    all_child_names = get_all_keys(schema.children, comm)
    for cn in all_child_names:
        if group is not None:
            new_group = group.require_group(cn)
        else:
            new_group = None

        child_schema = schema.children.get(cn, make_schema(cn, FileEntry.EMPTY))

        __allocate(child_schema, new_group, comm)


def get_column_allocation_metadata(column: Optional[ColumnWriter], comm: MPI.Comm):
    """
    Determine how to allocate the column. The most important thing this does is
    determine the overal shape. Keep in mind we have already done verification
    at this point to make sure everything is consistent.
    """
    strategy = None if column is None else column.combine_strategy
    strategies = list(filter(lambda strat: strat is not None, comm.allgather(strategy)))
    strategy = strategies[0]

    meta = None if column is None else (column.shape, column.dtype, column.attrs)
    all_meta = list(filter(lambda cm: cm is not None, comm.allgather(meta)))
    reference_meta = all_meta[0]

    if strategy == ColumnCombineStrategy.CONCAT:
        total_length = sum(m[0][0] for m in all_meta)
    else:
        total_length = reference_meta[0][0]

    # Note, we have already verified that all data types have the same
    # kind, so we simply need to pick the highest-precision one.
    all_dtypes = [m[1] for m in all_meta]
    all_dtypes.sort(key=lambda dt: dt.itemsize, reverse=True)

    shape = (total_length,) + reference_meta[0][1:]
    return shape, all_dtypes[0], all_meta[0][2]


def get_column_offset(column: Optional[ColumnWriter], comm: MPI.Comm):
    """
    Determine the offset for a given column on this rank.
    """
    if column is None or column.combine_strategy == ColumnCombineStrategy.SUM:
        length = 0
    else:
        length = len(column)
    all_lengths = comm.allgather(length)
    offsets = np.insert(np.cumsum(all_lengths), 0, 0)
    return offsets[comm.Get_rank()]


def __write_metadata(
    schema: Schema, group: Optional[h5py.File | h5py.Group], comm: MPI.Comm
):
    """
    Write metadata-only groups.
    """
    if schema.type == FileEntry.EMPTY:
        attrs = comm.allgather(None)
    else:
        attrs = comm.allgather(schema.attributes)
    attrs_to_write = list(filter(lambda at: at is not None, attrs))[0]
    if group is not None:
        for path, metadata in attrs_to_write.items():
            metadata_group = group.require_group(path)
            metadata_group.attrs.update(metadata)


def __allocate_column(
    name: str,
    column_writer: Optional[ColumnWriter],
    group: Optional[h5py.Group | h5py.File],
    comm: MPI.Comm,
) -> Optional[h5py.Dataset]:
    """
    Allocate a single column
    """

    shape, dtype, attrs = get_column_allocation_metadata(column_writer, comm)
    if group is not None:
        ds = group.create_dataset(name, shape=shape, dtype=dtype)
        ds.attrs.update(attrs)
        return ds
    return None


def __write_columns(
    schema: Schema,
    group: h5py.File | h5py.Group,
    offsets: dict,
    comm: Optional[MPI.Comm],
):
    all_column_names = get_all_keys(schema.columns, comm)
    for cn in all_column_names:
        writer = schema.columns.get(cn)
        ds = group[cn]
        offset = offsets[ds.name]
        assert ds is not None
        __write_column(writer, ds, offset, comm)


def __write_column(
    writer: Optional[ColumnWriter],
    ds: h5py.Dataset,
    offset: int,
    comm: Optional[MPI.Comm],
):
    strategy = None if writer is None else writer.combine_strategy
    if comm is not None:
        strategies = list(
            filter(lambda strat: strat is not None, comm.allgather(strategy))
        )
        strategy = strategies[0]

    if comm is not None:
        participating = comm.allgather(writer is not None)
        participating_ranks = [i for i in range(len(participating)) if participating[i]]
        group = comm.Get_group()
        new_group = group.Incl(participating_ranks)
        new_comm = comm.Create(new_group)
    else:
        new_comm = None

    match strategy:
        case ColumnCombineStrategy.CONCAT:
            if writer is not None:
                data = writer.get_data(new_comm)
                ds.write_direct(data, dest_sel=np.s_[offset : offset + len(data)])
        case ColumnCombineStrategy.SUM:
            if writer is None:
                data = np.zeros(ds.shape, ds.dtype)
            else:
                data = writer.get_data(new_comm)
            if comm is not None:
                data_to_write = comm.allreduce(data)
                ds[:] = data_to_write
            else:
                data += ds[:]
                ds[:] = data

    if new_comm is not None:
        assert new_group is not None
        new_comm.Free()
        new_group.Free()
    if comm is not None:
        comm.Barrier()
    ds.file.flush()



from __future__ import annotations

from functools import partial, reduce
from typing import TYPE_CHECKING, Any, Iterable, Optional

import numpy as np

from opencosmo.index import into_array

if TYPE_CHECKING:
    import opencosmo as oc

"""
A tale in 3 acts:

Act 1: Patrick creates a LinkHandler which holds a pointer to the index that determines
which rows in halo/galaxy properties corresponds to rows in particle datasets.

Act 2: Patrick unifies metadata handling in Datasets, making the link handler
unncessary. Instead, ephemeral particle datasets are created when requested.

Act 3: Patrick realizes this solution makes it impossible to cache things, particularly
very expensive computations that the user has created with evaluate. Patrick re-introduces
the LinkHandler, but it's better this time or something.

"""

LINK_ALIASES = {  # Left: Name in file, right: Name in collection
    "sodbighaloparticles_star_particles": "star_particles",
    "sodbighaloparticles_dm_particles": "dm_particles",
    "sodbighaloparticles_gravity_particles": "gravity_particles",
    "sodbighaloparticles_agn_particles": "agn_particles",
    "sodbighaloparticles_gas_particles": "gas_particles",
    "sod_profile": "halo_profiles",
    "galaxyproperties": "galaxy_properties",
    "galaxyparticles_star_particles": "star_particles",
}


def create_start_size(data, start_name, size_name):
    start = data.pop(start_name, None)
    size = data.pop(size_name, None)
    if start is None:
        return None
    valid = size > 0
    if isinstance(start, np.ndarray):
        return (start[valid], size[valid])
    if size == 0:
        return None
    return (np.atleast_1d(start), np.atleast_1d(size))


def create_idx(data, idx_name):
    idx = data.pop(idx_name, None)
    if idx is None:
        return None

    valid = idx >= 0

    if isinstance(idx, np.ndarray):
        return idx[valid]
    elif idx == -1:
        return None
    return np.atleast_1d(idx)


def make_links(keys, rename_galaxies=False):
    starts = list(filter(lambda key: "start" in key, keys))
    sizes = list(filter(lambda key: "size" in key, keys))
    idxs = list(filter(lambda key: "idx" in key, keys))

    starts = set(map(lambda key: key[:-6], starts))
    sizes = set(map(lambda key: key[:-5], sizes))
    idxs = set(map(lambda key: key[:-4], idxs))

    assert starts == sizes
    output = {}
    columns = {}
    for name in starts:
        output[LINK_ALIASES[name]] = partial(
            create_start_size, start_name=f"{name}_start", size_name=f"{name}_size"
        )
        columns[LINK_ALIASES[name]] = [f"{name}_start", f"{name}_size"]

    for name in idxs:
        output[LINK_ALIASES[name]] = partial(create_idx, idx_name=f"{name}_idx")
        columns[LINK_ALIASES[name]] = [f"{name}_idx"]

    if rename_galaxies and "galaxy_properties" in output:
        output["galaxies"] = output.pop("galaxy_properties")
        columns["galaxies"] = columns.pop("galaxy_properties")
    return output, columns


class LinkHandler:
    """
    This needs some explanation. We break the "don't mutate state" rule pretty hard here.

    When a StructureCollection is initialized, we build its linked datasets based on
    the metadata in the halo/galaxy properties which tells us which rows in the linked
    datasets belong to which halo/galaxy.

    When a StructureCollection is modified (e.g. by filtering) we filter the source
    dataset but DO NOT update the linked datasets, because doing so is a fairly
    expensive operation. Instead, we create a LinkHandler that knows which rows in
    the original halo/galaxy properties are included in the current version of the
    datasets.

    When data is requested, we perform the update if necessary. The StructureCollection
    passes in its current halo/galaxy properties and the un-updated linked datasets. We know
    for sure that the rows in the current halo/galaxy properties is a strict subset of the rows
    in the original halo/galaxy properties. We determine where the overlaps are, and then update
    the datasets accordingly.

    The original version of the StructureCollection held the entire, unmodified linked datasets
    and created the version with the correct rows when the data was requested. While this was very clean,
    it made it impossible to cache data because the linked datasets that actually got returned to the
    user were ephemeral.

    We could re-build the datasets each time we create a new StructureCollection, but this is quite slow.
    Converting from that approach to this approach shaved 25% off the runtime of one of my tests.
    """

    def __init__(
        self,
        links,
        columns,
        derived_from: Optional[oc.Dataset],
    ):
        self.__derived_from = derived_from
        self.links = links
        self.columns = columns

    @classmethod
    def from_link_names(cls, names: Iterable[str], rename_galaxies=False):
        links, columns = make_links(names, rename_galaxies)
        return LinkHandler(links, columns, None)

    def parse(self, data: dict[str, Any]):
        output = {}
        for name, handler in self.links.items():
            result = handler(data)
            if result is not None:
                output[name] = result
        return output

    def prep_datasets(self, source: oc.Dataset, datasets: dict[str, oc.Dataset]):
        """
        Called once when a datasets are opened for the first time. Downstream
        versions always use rebuild_datsets
        """

        all_columns: list[str] = reduce(
            lambda acc, ds: acc + self.columns[ds], datasets.keys(), []
        )
        meta = source.get_metadata(all_columns)
        indices = self.parse(meta)
        new_datasets = datasets
        for name, index in indices.items():
            new_datasets[name] = new_datasets[name].take_rows(index)
        return new_datasets

    def make_derived(self, source: oc.Dataset):
        """
        Because the library encourages performing several operations in a row before requesting data,
        it is a bad idea to re-build the datasets every time someone performs an operation. Instead,
        we just hold a copy of the source dataset the last time the datasets were rebuilt,
        which allows us to perform the rebuild whenver data is actually requsted.
        """
        derived_from = self.__derived_from
        if self.__derived_from is None:
            derived_from = source

        return LinkHandler(self.links, self.columns, derived_from)

    def rebuild_datasets(
        self,
        new_source: oc.Dataset,
        datasets: dict[str, oc.Dataset],
    ):
        """
        We have a few guarantees here:
        1. The rows in new_source is a strict subset of the rows in source
        2. The rows in both are unique

        What is NOT guaranteed:
        1. The rows are sorted
        """
        if self.__derived_from is None:
            return datasets
        original_index = into_array(self.__derived_from.index)
        new_index = into_array(new_source.index)

        _, index_into_original, index_into_new = np.intersect1d(
            original_index, new_index, assume_unique=True, return_indices=True
        )
        all_columns: list[str] = reduce(
            lambda acc, ds: acc + self.columns[ds], datasets.keys(), []
        )
        index_into_original = index_into_original[np.argsort(index_into_new)]
        metadata = self.__derived_from.get_metadata(all_columns)
        new_datasets = {}

        for name, dataset in datasets.items():
            if len(self.columns[name]) == 1:
                metadata_column = metadata[self.columns[name][0]]
                new_datasets[name] = rebuild_row_index(
                    new_source, dataset, metadata_column, index_into_original
                )
            else:
                size_column = [name for name in self.columns[name] if "size" in name]
                assert len(size_column) == 1
                size_column_name = size_column[0]
                size_column_data = metadata[size_column_name]
                new_datasets[name] = rebuild_chunk_index(
                    new_source, dataset, size_column_data, index_into_original
                )
        return new_datasets

    def resort(self, source: oc.Dataset, datasets: dict[str, oc.Dataset]):
        """
        Data is always written in its original order, whether or not it has been sorted.
        This is to preserve the spatial index. However, when linked datasets are rebuilt
        they are rebuilt in the sorted order. This method re-sorts them based on the
        index from the original data.
        """
        all_columns: list[str] = reduce(
            lambda acc, ds: acc + self.columns[ds], datasets.keys(), []
        )
        all_columns = list(
            filter(lambda name: "idx" in name or "size" in name, all_columns)
        )

        sort_index = np.argsort(into_array(source.index))

        if np.all(sort_index[1:] >= sort_index[:-1]):
            # Already sorted. Carry on!
            return datasets

        meta = source.get_metadata(all_columns)
        output = {}
        for name, dataset in datasets.items():
            if len(self.columns[name]) == 1:
                valid_rows = meta[self.columns[name][0]] >= 0
                new_dataset = dataset.take_rows(sort_index[valid_rows])
            else:
                size_column = [name for name in self.columns[name] if "size" in name]
                assert len(size_column) == 1
                size_column_data = meta[size_column[0]]
                chunk_boundaries = np.zeros(len(size_column_data) + 1, dtype=int)
                _ = np.cumsum(size_column_data, out=chunk_boundaries[1:])
                starts = chunk_boundaries[sort_index]
                sizes = size_column_data[sort_index]
                valid = sizes > 0
                idx = (starts[valid], sizes[valid])
                new_dataset = dataset.take_rows(idx)
            output[name] = new_dataset
        return output


def rebuild_row_index(
    new_source: oc.Dataset,
    dataset: oc.Dataset,
    original_metadata_column: np.ndarray,
    index_into_original: np.ndarray,
):
    valid_rows = original_metadata_column >= 0
    index = np.full(len(original_metadata_column), -1, dtype=int)
    index[valid_rows] = np.arange(0, sum(valid_rows))
    index_to_take = index[index_into_original]
    index_to_take = index_to_take[index_to_take >= 0]

    return dataset.take_rows(index_to_take)


def rebuild_chunk_index(
    new_source: oc.Dataset,
    dataset: oc.Dataset,
    original_size_column: np.ndarray,
    index_into_original: np.ndarray,
):
    chunk_boundaries = np.zeros(len(original_size_column) + 1, dtype=int)
    _ = np.cumsum(original_size_column, out=chunk_boundaries[1:])
    valid_rows = original_size_column[index_into_original] > 0

    starts = chunk_boundaries[index_into_original[valid_rows]]
    sizes = original_size_column[index_into_original[valid_rows]]

    ds = dataset.take_rows((starts, sizes))
    return ds



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import h5py

    from opencosmo.io.schema import Schema


def allocate(group: h5py.File | h5py.Group, schema: Schema):
    for column_name, column_writer in schema.columns.items():
        group.require_dataset(column_name, column_writer.shape, column_writer.dtype)
    for child_name, child_schema in schema.children.items():
        child_group = group.require_group(child_name)
        allocate(child_group, child_schema)


def write_columns(group: h5py.File | h5py.Group, schema: Schema):
    for column_path, column_writer in schema.columns.items():
        group[column_path][:] = column_writer.data
        group[column_path].attrs.update(column_writer.attrs)
    for child_name, child_schema in schema.children.items():
        write_columns(group[child_name], child_schema)


def write_metadata(group: h5py.File | h5py.Group, schema: Schema):
    for path, metadata in schema.attributes.items():
        metadata_group = group.require_group(path)
        metadata_group.attrs.update(metadata)

    for child_name, child_schema in schema.children.items():
        write_metadata(group[child_name], child_schema)



from __future__ import annotations

from inspect import Parameter, signature
from itertools import chain
from typing import TYPE_CHECKING, Any, Callable, Iterable, Sequence

from astropy.units import Quantity

from opencosmo.column.column import EvaluatedColumn
from opencosmo.column.evaluate import EvaluateStrategy, do_first_evaluation
from opencosmo.evaluate import (
    insert_data,
    make_output_from_first_values,
)

if TYPE_CHECKING:
    import numpy as np

    from opencosmo import Dataset

"""
Although the user-facing name for this operation is "evaluate", the pattern 
we are using here is known as a "visitor."
"""


def visit_dataset(
    function: Callable,
    strategy: str,
    format: str,
    evaluator_kwargs: dict[str, Any],
    dataset: Dataset,
):
    column = verify_for_lazy_evaluation(
        function,
        strategy,
        format,
        evaluator_kwargs,
        dataset,
        skip_evaluation_check=True,
    )

    data = dataset.select(column.requires).get_data(output=format)
    try:
        data = dict(data)
    except TypeError:
        data = {column.requires.pop(): data}
    output = column.evaluate(data, dataset.index)
    if not isinstance(output, dict):
        assert len(column.produces) == 1
        output = {column.produces.pop(): output}
    return output


def verify_for_lazy_evaluation(
    func: Callable,
    strategy: str,
    format: str,
    evaluator_kwargs: dict[str, Any],
    dataset: Dataset,
    allow_none=False,
    skip_evaluation_check=False,
) -> EvaluatedColumn:
    """
    Verify the function behaves correctly and determine the names of its output columns.
    """
    __verify(func, dataset.columns, evaluator_kwargs.keys())
    sig = signature(func)
    required_arguments = filter(
        lambda param: param.default == Parameter.empty, sig.parameters.values()
    )
    required_argument_names = set(map(lambda param: param.name, required_arguments))
    required_columns = required_argument_names.difference(evaluator_kwargs.keys())

    if diff := required_columns.difference(dataset.columns):
        raise ValueError(
            f"Function expects columns {diff} which are not in the dataset"
        )
    dataset = dataset.select(required_columns)
    if skip_evaluation_check:
        first_values = None
        eval_strategy = EvaluateStrategy(strategy)
    else:
        first_values, eval_strategy = do_first_evaluation(
            func, strategy, format, evaluator_kwargs, dataset
        )
        if first_values is None and not allow_none:
            raise ValueError(
                "Cannot insert values from an evaluate function that returns None!"
            )

    if isinstance(first_values, dict):
        produces = set(first_values.keys())
    else:
        produces = {func.__name__}
    column = EvaluatedColumn(
        func, required_columns, produces, format, eval_strategy, **evaluator_kwargs
    )
    return column


def __visit_rows_in_dataset(
    function: Callable,
    dataset: Dataset,
    format: str,
    kwargs: dict[str, Any] = {},
    iterable_kwargs: dict[str, Sequence] = {},
):
    first_row_values = dict(dataset.take(1, at="start").get_data())
    first_row_kwargs = kwargs | {name: arr[0] for name, arr in iterable_kwargs.items()}
    storage = __make_output(function, first_row_values | first_row_kwargs, len(dataset))
    for i, row in enumerate(dataset.rows(include_units=format == "astropy")):
        if i == 0:
            continue
        iter_kwargs = {name: arr[i] for name, arr in iterable_kwargs.items()}
        output = function(**row, **kwargs, **iter_kwargs)
        if storage is not None:
            insert_data(storage, i, output)
    return storage


def __visit_rows_in_data(
    function: Callable,
    data: dict[str, np.ndarray],
    format="astropy",
    kwargs: dict[str, Any] = {},
    iterable_kwargs: dict[str, np.ndarray] = {},
):
    data = {key: d for key, d in data.items() if key in signature(function).parameters}
    first_row_data = {name: arr[0] for name, arr in data.items()}
    first_row_kwargs = kwargs | {name: arr[0] for name, arr in iterable_kwargs.items()}
    n_rows = len(next(iter(data.values())))
    storage = __make_output(function, first_row_data | first_row_kwargs, n_rows)
    if format == "numpy":
        data = {
            key: arr.value if isinstance(arr, Quantity) else arr
            for key, arr in data.items()
        }

    for i in range(1, n_rows):
        row = {
            name: arr[i] for name, arr in chain(data.items(), iterable_kwargs.items())
        }
        output = function(**row, **kwargs)
        if storage is not None:
            insert_data(storage, i, output)
    return storage


def __make_output(
    function: Callable,
    first_input_values: dict[str, Any],
    n_rows: int,
) -> dict | None:
    first_values = function(**first_input_values)
    if first_values is None:
        return None
    if not isinstance(first_values, dict):
        name = function.__name__
        first_values = {name: first_values}

    return make_output_from_first_values(first_values, n_rows)


def __visit_vectorize(
    function: Callable,
    data: dict[str, Iterable] | Iterable,
    evaluator_kwargs: dict[str, Any] = {},
):
    pars = signature(function).parameters

    if not isinstance(data, dict) or (len(data) > 1 and len(pars) == 1):
        return function(data, **evaluator_kwargs)

    input_data = {pname: data[pname] for pname in pars if pname in data}

    return function(**input_data, **evaluator_kwargs)


def __verify(
    function: Callable, data_columns: Iterable[str], kwarg_names: Iterable[str]
):
    function_signature = signature(function)
    required_parameters = set()
    for name, parameter in function_signature.parameters.items():
        if parameter.default == Parameter.empty:
            required_parameters.add(name)

    missing = required_parameters.difference(data_columns).difference(kwarg_names)
    if not missing:
        return required_parameters.intersection(data_columns)
    elif len(missing) > 1:
        raise ValueError(
            f"All inputs to the function must either be column names or passed as keyword arguments! Found unknown input(s) {','.join(missing)}"
        )



from __future__ import annotations

from inspect import Parameter, signature
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence

import numpy as np
from astropy.units import Quantity  # type: ignore

from opencosmo import dataset as ds
from opencosmo.evaluate import (
    insert_data,
    make_output_from_first_values,
    prepare_kwargs,
)

if TYPE_CHECKING:
    from opencosmo import StructureCollection


def verify_evaluate_on_collection(
    function: Callable,
    collection: StructureCollection,
    evaluate_kwargs: dict[str, Any],
    dataset: Optional[str],
):
    # Case 1/3
    function_signature = signature(function)
    function_arg_names = set(function_signature.parameters.keys())
    datasets_in_collection = set(collection.keys())
    if not (
        requested_datasets := function_arg_names.intersection(datasets_in_collection)
    ):
        raise ValueError(
            "Your function should take the names of some of the datasets in this collection as arguments!"
        )
    elif dataset is not None and dataset not in requested_datasets:
        raise ValueError(
            "If you pass an argument to 'dataset', your function must take in at least one column from that dataset"
        )

    required_parameters = {
        name
        for name, par in function_signature.parameters.items()
        if par.default == Parameter.empty
    }
    if missing := required_parameters.difference(datasets_in_collection).difference(
        evaluate_kwargs.keys()
    ):
        raise ValueError(
            f"Your function has required arguments {missing}, but you didn't provide them!"
        )

    spec = {name: evaluate_kwargs.pop(name, None) for name in requested_datasets}
    return spec, evaluate_kwargs


def visit_structure_collection_eagerly(
    function: Callable,
    collection: StructureCollection,
    format: str = "astropy",
    dataset: Optional[str] = None,
    evaluate_kwargs: dict[str, Any] = {},
    insert: bool = True,
):
    spec, kwargs = verify_evaluate_on_collection(
        function, collection, evaluate_kwargs, dataset
    )

    to_visit = __prepare_collection(spec, collection)

    if dataset is None:
        return evaluate_into_properties(function, to_visit, format, kwargs, insert)
    else:
        return evaluate_into_dataset(
            function, to_visit, format, kwargs, dataset, insert
        )


def evaluate_into_properties(
    function: Callable,
    collection: StructureCollection,
    format: str,
    kwargs: dict[str, Any],
    insert: bool,
):
    kwargs, iterable_kwargs = prepare_kwargs(len(collection), kwargs)

    storage = __make_output(
        function, collection, format, kwargs, iterable_kwargs, insert
    )
    for i, structure in enumerate(collection.objects()):
        if i == 0:
            continue
        iterable_kwarg_values = {name: arr[i] for name, arr in iterable_kwargs.items()}
        input_structure = __make_input(structure, format)

        output = function(**input_structure, **kwargs, **iterable_kwarg_values)
        if storage is not None:
            insert_data(storage, i, output)

    return storage


def evaluate_into_dataset(
    function: Callable,
    collection: StructureCollection,
    format: str,
    kwargs: dict[str, Any],
    dataset: str,
    insert: bool,
):
    kwargs, iterable_kwargs = prepare_kwargs(len(collection[dataset]), kwargs)
    storage = __make_chunked_output(
        function, collection, dataset, format, kwargs, iterable_kwargs
    )

    for i, structure in enumerate(collection.objects()):
        if i == 0:
            continue
        iterable_kwarg_values = {name: arr[i] for name, arr in iterable_kwargs.items()}
        input_structure = __make_input(structure, format)

        output = function(**input_structure, **kwargs, **iterable_kwarg_values)
        if storage is not None:
            for name, output_arr in output:
                storage[name].append(output_arr)

    if storage is None:
        return
    output_data = {name: np.concatenate(data) for name, data in storage.items()}
    return output_data


def __make_input(structure: dict, format: str = "astropy"):
    values = {}
    for name, element in structure.items():
        if isinstance(element, dict):
            values[name] = __make_input(element, format)
        elif isinstance(element, ds.Dataset):
            data = element.get_data(format)
            values[name] = data
        elif isinstance(element, Quantity) and format == "numpy":
            values[name] = element.value
        else:
            values[name] = element
    return values


def __make_output(
    function: Callable,
    collection: StructureCollection,
    format: str = "astropy",
    kwargs: dict[str, Any] = {},
    iterable_kwargs: dict[str, Sequence] = {},
    insert: bool = True,
) -> dict | None:
    first_structure = next(collection.take(1, at="start").objects())
    first_input = __make_input(first_structure, format)
    first_values = function(
        **first_input,
        **kwargs,
        **{name: arr[0] for name, arr in iterable_kwargs.items()},
    )
    if first_values is None and insert:
        raise ValueError(
            "You asked to insert these values, but your function returns None!"
        )
    elif first_values is None:
        return None
    if not isinstance(first_values, dict):
        name = function.__name__
        first_values = {name: first_values}
    n_rows = len(collection)
    return make_output_from_first_values(first_values, n_rows)


def __make_chunked_output(
    function: Callable,
    collection: StructureCollection,
    dataset: str,
    format: str = "astropy",
    kwargs: dict[str, Any] = {},
    iterable_kwargs: dict[str, Sequence] = {},
    insert: bool = True,
) -> dict | None:
    first_structure = collection.take(1, at="start").objects()
    expected_length = len(first_structure[dataset])
    first_structure_data = next(iter(first_structure.objects()))

    first_input = __make_input(first_structure_data, format)
    first_values = function(
        **first_input,
        **kwargs,
        **{name: arr[0] for name, arr in iterable_kwargs.items()},
    )
    if first_values is None and insert:
        raise ValueError(
            "You asked to insert these values, but your function returns None!"
        )
    elif first_values is None:
        return None
    if not isinstance(first_values, dict):
        name = function.__name__
        first_values = {name: first_values}
    if any(len(fv) != expected_length for fv in first_values.values()):
        raise ValueError(
            "If you pass a `dataset` argument, your function should output an array with the same length as that dataset"
        )
    return {name: [fv] for name, fv in first_values.items()}


def __prepare_collection(
    spec: dict[str, Optional[list[str]]], collection: StructureCollection
) -> StructureCollection:
    collection = collection.with_datasets(list(spec.keys()))
    selections = {ds_name: cols for ds_name, cols in spec.items() if cols is not None}
    collection = collection.select(**selections)
    return collection



import numpy as np
from numpy.typing import NDArray

from .build import concatenate, empty, from_size, single_chunk
from .get import get_data
from .in_range import n_in_range
from .mask import into_array, mask
from .project import project
from .take import take
from .unary import get_length, get_range

SimpleIndex = NDArray[np.int_]
ChunkedIndex = tuple[NDArray[np.int_], NDArray[np.int_]]
DataIndex = SimpleIndex | ChunkedIndex


__all__ = [
    "DataIndex",
    "SimpleIndex",
    "ChunkedIndex",
    "empty",
    "from_size",
    "single_chunk",
    "concatenate",
    "get_data",
    "get_length",
    "get_range",
    "into_array",
    "mask",
    "n_in_range",
    "project",
    "take",
    "get_length",
    "get_range",
]



from __future__ import annotations

from functools import reduce
from itertools import product
from typing import TYPE_CHECKING, Mapping, Optional

import rustworkx as rx

if TYPE_CHECKING:
    import astropy.units as u
    import numpy as np

    from opencosmo.column.cache import ColumnCache
    from opencosmo.column.column import ConstructedColumn
    from opencosmo.dataset.handler import Hdf5Handler
    from opencosmo.index import DataIndex
    from opencosmo.units.handler import UnitHandler


def build_dependency_graph(
    derived_columns: Mapping[str, ConstructedColumn],
    names_to_keep: Optional[set[str]] = None,
):
    dependency_graph = rx.PyDiGraph()
    all_requires: set[str] = reduce(
        lambda known, dc: known.union(dc.requires), derived_columns.values(), set()
    )
    nodeidx = dependency_graph.add_nodes_from(all_requires)
    nodemap = {name: idx for (name, idx) in zip(all_requires, nodeidx)}

    for target, derived_column in derived_columns.items():
        requires = derived_column.requires
        produces = derived_column.produces
        if produces is None:
            produces = set((target,))
        to_add = list(filter(lambda p: p not in nodemap, produces))
        new_map = dependency_graph.add_nodes_from(to_add)
        nodemap.update({name: idx for (name, idx) in zip(to_add, new_map)})

        requires_idx = tuple(nodemap[r] for r in requires)
        produces_idx = tuple(nodemap[r] for r in produces)

        dependency_graph.add_edges_from_no_data(product(requires_idx, produces_idx))

    if names_to_keep is not None:
        nodes_to_keep = reduce(
            lambda acc, name: acc.union(rx.ancestors(dependency_graph, nodemap[name])),
            names_to_keep,
            {nodemap[name] for name in names_to_keep},
        )
        names_to_keep = {dependency_graph[n] for n in nodes_to_keep}
        dependency_graph = dependency_graph.subgraph(list(nodes_to_keep))
        derived_columns = {
            name: dc for name, dc in derived_columns.items() if name in names_to_keep
        }

    return dependency_graph


def replace_multi_producers(
    graph: rx.PyDiGraph, derived_columns: Mapping[str, ConstructedColumn]
):
    """
    Some derived columns actually produce multiple outputs. At this stage, the dependency
    graph is working solely with actual column names, meaning if any of those columns is
    produced by one of these "multi-produces" they will not be in the derived_columns
    dictionary and therefore cannot be instantiated. This function replaces such
    columns with the name of the derived_column that produces them.
    """

    node_map = {name: i for i, name in enumerate(graph.nodes())}
    missing = set(derived_columns.keys()).difference(node_map.keys())
    if not missing:
        return graph
    for missing_column in missing:
        missing_column_produces = derived_columns[missing_column].produces
        if missing_column_produces is None:
            continue
        outputs = [
            node_map[name] for name in missing_column_produces if name in node_map
        ]
        graph.contract_nodes(outputs, missing_column)
    return graph


def validate_derived_columns(
    derived_columns: dict[str, ConstructedColumn],
    known_raw_columns: set[str],
    units: dict[str, u.Unit],
):
    """
    Validate the network of derived columns. This
    """
    dependency_graph = build_dependency_graph(derived_columns)
    if cycle := rx.digraph_find_cycle(dependency_graph):
        all_nodes: set[int] = reduce(
            lambda known, edge: known.union(edge), cycle, set()
        )
        names = [dependency_graph[i] for i in all_nodes]
        raise ValueError(
            f"Found derived columns that depend on each other! Columns: {names}"
        )

    sources = set(
        filter(
            lambda i: not dependency_graph.in_degree(i),
            range(dependency_graph.num_nodes()),
        )
    )
    source_names = map(lambda i: dependency_graph[i], sources)
    if missing := set(source_names).difference(known_raw_columns):
        raise ValueError(f"Tried to derive columns from unknown columns: {missing}")

    dependency_graph = replace_multi_producers(dependency_graph, derived_columns)
    validate_dependency_graph(
        dependency_graph, known_raw_columns, set(derived_columns.keys())
    )

    return validate_derived_units(dependency_graph, derived_columns, units)


def validate_dependency_graph(
    dependency_graph: rx.PyDiGraph,
    known_raw_columns: set[str],
    derived_columns: set[str],
):
    expected = set(dependency_graph.nodes())
    known = known_raw_columns.union(derived_columns)
    missing = expected.difference(known)
    assert len(missing) == 0


def validate_derived_units(
    dependency_graph: rx.PyDiGraph,
    derived_columns: dict[str, ConstructedColumn],
    units: dict[str, u.Unit],
):
    output_units: dict[str, Optional[u.Unit]] = {}
    for node in rx.topological_sort(dependency_graph):
        node_name = dependency_graph[node]
        if node_name in units:
            continue
        new_units = derived_columns[node_name].get_units(units)
        if not isinstance(new_units, dict):
            new_units = {node_name: new_units}
        units |= new_units
        output_units |= new_units
    return output_units


def build_derived_columns(
    all_derived_columns: dict[str, ConstructedColumn],
    derived_columns_to_get: set[str],
    cache: ColumnCache,
    hdf5_handler: Hdf5Handler,
    unit_handler: UnitHandler,
    unit_kwargs: dict,
    index: DataIndex,
) -> dict[str, np.ndarray]:
    """
    Build any derived columns that are present in this dataset. Also returns any columns that
    had to be instantiated in order to build these derived columns.
    """
    if not derived_columns_to_get:
        return {}

    column_names: set[str] = reduce(
        lambda known, dc: known.union(dc[1].produces)
        if dc[1].produces is not None
        else known.union((dc[0],)),
        all_derived_columns.items(),
        set(),
    )

    dependency_graph = build_dependency_graph(
        all_derived_columns, derived_columns_to_get
    )
    cached_data = cache.get_columns(dependency_graph.nodes())
    cached_data |= unit_handler.apply_unit_conversions(cached_data, unit_kwargs)

    additional_derived = column_names.difference(cached_data.keys())

    if not additional_derived:
        return cached_data

    columns_to_fetch = (
        set(dependency_graph.nodes())
        .intersection(hdf5_handler.columns)
        .difference(cached_data.keys())
    )

    raw_data = hdf5_handler.get_data(columns_to_fetch)
    data = cached_data | unit_handler.apply_units(raw_data, unit_kwargs)

    dependency_graph = replace_multi_producers(dependency_graph, all_derived_columns)
    new_derived: dict[str, np.ndarray] = {}

    for colidx in rx.topological_sort(dependency_graph):
        colname = dependency_graph[colidx]
        if colname in data:
            continue
        derived_column = all_derived_columns[colname]
        produces = derived_column.produces
        if produces is None:
            produces = set((colname,))
        if all(name in data for name in produces):
            continue
        output = derived_column.evaluate(data, index)
        if isinstance(output, dict):
            data |= output
            new_derived |= output
        else:
            data[colname] = output
            new_derived[colname] = output

    if new_derived:
        cache.add_data(new_derived)

    return data | new_derived



from __future__ import annotations

from functools import cached_property
from itertools import chain
from typing import TYPE_CHECKING, Iterable, Optional

import numpy as np

from opencosmo.index import (
    SimpleIndex,
    from_size,
    get_data,
    get_length,
    into_array,
    take,
)
from opencosmo.io.schema import FileEntry, make_schema
from opencosmo.io.writer import (
    ColumnWriter,
)

if TYPE_CHECKING:
    import h5py

    from opencosmo.header import OpenCosmoHeader
    from opencosmo.index import DataIndex
    from opencosmo.io.schema import Schema


class Hdf5Handler:
    """
    Handler for opencosmo.Dataset
    """

    def __init__(
        self,
        group: h5py.Group,
        index: DataIndex,
        metadata_group: Optional[h5py.Group],
        load_conditions: Optional[dict[str, bool]] = None,
    ):
        self.__index = index
        self.__data_group = group
        self.__metadata_group = metadata_group
        self.__in_memory = group.file.driver == "core"
        self.__load_conditions = load_conditions

    @classmethod
    def from_group(
        cls,
        group: h5py.Group,
        index: Optional[DataIndex] = None,
        metadata_group: Optional[h5py.Group] = None,
        load_conditions: Optional[dict[str, bool]] = None,
    ):
        if not group.name.endswith("data"):
            raise ValueError("Expected a data group")
        lengths = set(len(ds) for ds in group.values())
        if len(lengths) > 1:
            raise ValueError("Not all columns are the same length!")

        if index is None:
            index = from_size(lengths.pop())

        colnames = group.keys()
        if metadata_group is not None:
            colnames = chain(colnames, metadata_group.keys())

        return Hdf5Handler(group, index, metadata_group, load_conditions)

    @property
    def in_memory(self) -> bool:
        return self.__in_memory

    @property
    def load_conditions(self) -> Optional[dict[str, bool]]:
        return self.__load_conditions

    def take(self, other: DataIndex, sorted: Optional[np.ndarray] = None):
        if len(other) == 0:
            return Hdf5Handler(
                self.__data_group, other, self.__metadata_group, self.__load_conditions
            )

        if sorted is not None:
            return self.__take_sorted(other, sorted)

        new_index = take(self.__index, other)
        return Hdf5Handler(
            self.__data_group, new_index, self.__metadata_group, self.__load_conditions
        )

    def __take_sorted(self, other: DataIndex, sorted: np.ndarray):
        if get_length(sorted) != get_length(self.__index):
            raise ValueError("Sorted index has the wrong length!")
        new_indices = get_data(other, sorted)

        new_raw_index = into_array(self.__index)[new_indices]
        new_index = np.sort(new_raw_index)

        return Hdf5Handler(
            self.__data_group, new_index, self.__metadata_group, self.__load_conditions
        )

    @property
    def data(self):
        return self.__data_group

    @property
    def index(self):
        return self.__index

    @property
    def columns(self):
        return self.__data_group.keys()

    @property
    def metadata_columns(self):
        if self.__metadata_group is None:
            return []
        return self.__metadata_group.keys()

    @cached_property
    def descriptions(self):
        return {
            colname: column.attrs.get("description")
            for colname, column in self.__data_group.items()
        }

    def mask(self, mask):
        idx = SimpleIndex(np.where(mask)[0])
        return self.take(idx)

    def __len__(self) -> int:
        first_column_name = next(iter(self.__data_group.keys()))
        return self.__data_group[first_column_name].shape[0]

    def __enter__(self):
        return self

    def __exit__(self, *exec_details):
        self.__data_group = None
        return self.__file.close()

    def make_schema(
        self,
        columns: Iterable[str],
        header: Optional[OpenCosmoHeader] = None,
    ) -> tuple[Schema, Optional[Schema]]:
        column_writers = {}
        for column_name in columns:
            column_writers[column_name] = ColumnWriter.from_h5_dataset(
                self.__data_group[column_name],
                self.__index,
                attrs=dict(self.__data_group[column_name].attrs),
            )

        data_schema = make_schema("data", FileEntry.COLUMNS, columns=column_writers)

        metadata_schema = None

        if self.metadata_columns:
            assert self.__metadata_group is not None
            group_name = self.__metadata_group.name.split("/")[-1]
            metadata_writers = {}
            for column_name in self.metadata_columns:
                metadata_writers[column_name] = ColumnWriter.from_h5_dataset(
                    self.__metadata_group[column_name],
                    self.__index,
                    attrs=dict(self.__metadata_group[column_name].attrs),
                )
            metadata_schema = make_schema(
                group_name, FileEntry.COLUMNS, columns=metadata_writers
            )
        return data_schema, metadata_schema

    def get_data(self, columns: Iterable[str]) -> dict[str, np.ndarray]:
        """ """
        if self.__data_group is None:
            raise ValueError("This file has already been closed")
        data = {}

        for colname in columns:
            data[colname] = get_data(self.__data_group[colname], self.__index)
        # Ensure order is preserved
        return {name: data[name] for name in columns}

    def get_metadata(self, columns: Iterable[str]) -> Optional[dict[str, np.ndarray]]:
        if self.__metadata_group is None:
            return None
        if not columns:
            columns = self.metadata_columns

        data = {}
        for colname in columns:
            data[colname] = get_data(self.__metadata_group[colname], self.__index)

        return data

    def take_range(self, start: int, end: int, indices: np.ndarray) -> np.ndarray:
        if start < 0 or end > len(indices):
            raise ValueError("Indices out of range")
        return indices[start:end]



from __future__ import annotations

import operator as op
from copy import copy
from functools import cached_property, partial, partialmethod
from inspect import signature
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Optional,
    Protocol,
    Self,
    Union,
)

import astropy.units as u  # type: ignore
import numpy as np
from astropy import table  # type: ignore

from opencosmo.column.evaluate import (
    EvaluateStrategy,
    evaluate_chunks,
    evaluate_rows,
    evaluate_vectorized,
)
from opencosmo.units import UnitsError

if TYPE_CHECKING:
    from opencosmo import Dataset
    from opencosmo.index import DataIndex

Comparison = Callable[[float, float], bool]


def col(column_name: str) -> Column:
    """
    Create a reference to a column with a given name. These references can be combined
    to produce new columns or express queries that operate on the values in a given
    dataset. For example:

    .. code-block:: python

        import opencosmo as oc
        ds = oc.open("haloproperties.hdf5")
        query = oc.col("fof_halo_mass") > 1e14
        px = oc.col("fof_halo_mass") * oc.col("fof_halo_com_vx")
        ds = ds.with_new_columns(fof_halo_com_px = px).filter(query)

    For more advanced usage, see :doc:`cols`

    """
    return Column(column_name)


ColumnOrScalar = Union["Column", "DerivedColumn", int, float]


def _log10(
    left: np.ndarray | u.Unit,
    right: None,
    unit_container: u.LogUnit,
):
    vals = left
    unit = None
    if isinstance(left, u.UnitBase):
        return unit_container(left)

    elif isinstance(left, u.Quantity):
        vals = left.value
        unit = left.unit
        if isinstance(unit, u.LogUnit):
            raise ValueError("Cannot take the log of a log unit!")

    new_vals = np.log10(vals)
    if unit is not None:
        return new_vals * unit_container(unit)
    return new_vals


def _exp10(
    left: np.ndarray | u.Unit,
    right: None,
    expected_unit_container: u.LogUnit,
):
    vals = left
    unit = None
    if isinstance(left, u.LogUnit):
        if not isinstance(left, expected_unit_container):
            raise ValueError(
                f"Expected a unit of type {expected_unit_container}, found {type(left)}"
            )
        return left.physical_unit

    elif isinstance(left, u.Quantity):
        vals = left.value
        unit = left.unit
        if not isinstance(unit, u.LogUnit):
            raise ValueError(
                "Can only raise 10 to a unitful value if the unit is logarithmic"
            )
        if not isinstance(unit, expected_unit_container):
            raise ValueError(
                f"Expected a unit of type {expected_unit_container}, found {type(left)}"
            )

    new_vals = 10**vals
    if unit is not None:
        return new_vals * unit.physical_unit
    return new_vals


def _sqrt(left: np.ndarray | u.Unit, right: None):
    return left**0.5


class Column:
    """
    Represents a reference to a column with a given name. Column reference
    are created independently of the datasets that actually contain data.
    You should not create this class directly, instead use :py:meth:`opencosmo.col`.

    Columns can be combined, and support comparison operators for masking datasets.

    Combinations:

        - Basic arithmetic with +, -, \\*, and /
        - Powers with :code:`\\*\\*`, and :code:`column.sqrt()`
        - log and exponentiation with :code:`column.log10()` and :code:`column.exp10()`

    Comparison operators:

        - Arithmetic comparisons such as <, <=, >, ==, !=
        - Membership with :code:`column.isin`

    In general, combinations of columns produce a :code:`DerivedColumn`, which can be treated
    the exact same was as basic Columns.

    For example, to compute the x-component of a halo's momentum, and then filter out
    halos below a certain value of that momentum

    .. code-block:: python

        import opencosmo as oc

        dataset = oc.open("haloproperties.hdf5")
        halo_px = oc.col("fof_halo_mass") * oc.col("fof_halo_com_vx")
        dataset = dataset.with_new_columns(fof_halo_com_px = halo_px)

        min_momentum_filter = oc.col("fof_halo_com_px) > 10**14
        dataset = dataset.filter(min_momentum_filter)

    """

    def __init__(self, column_name: str):
        self.column_name = column_name

    # mypy doesn't reason about eq and neq correctly
    def __eq__(self, other: float | u.Quantity) -> ColumnMask:  # type: ignore
        return ColumnMask(self.column_name, other, op.eq)

    def __ne__(self, other: float | u.Quantity) -> ColumnMask:  # type: ignore
        return ColumnMask(self.column_name, other, op.ne)

    def __gt__(self, other: float | u.Quantity) -> ColumnMask:
        return ColumnMask(self.column_name, other, op.gt)

    def __ge__(self, other: float | u.Quantity) -> ColumnMask:
        return ColumnMask(self.column_name, other, op.ge)

    def __lt__(self, other: float | u.Quantity) -> ColumnMask:
        return ColumnMask(self.column_name, other, op.lt)

    def __le__(self, other: float | u.Quantity) -> ColumnMask:
        return ColumnMask(self.column_name, other, op.le)

    def isin(self, other: Iterable[float | u.Quantity]) -> ColumnMask:
        return ColumnMask(self.column_name, other, np.isin)

    def __rmul__(self, other: Any) -> DerivedColumn:
        match other:
            case int() | float():
                return self * other
            case _:
                return NotImplemented

    def __mul__(self, other: Any) -> DerivedColumn:
        match other:
            case int() | float() | Column():
                return DerivedColumn(self, other, op.mul)
            case _:
                return NotImplemented

    def __rtruediv__(self, other: Any) -> DerivedColumn:
        match other:
            case int() | float():
                return DerivedColumn(other, self, op.truediv)
            case _:
                return NotImplemented

    def __truediv__(self, other: Any) -> DerivedColumn:
        match other:
            case int() | float() | Column():
                return DerivedColumn(self, other, op.truediv)
            case _:
                return NotImplemented

    def __pow__(self, other: Any) -> DerivedColumn:
        match other:
            case int() | float():
                return DerivedColumn(self, other, op.pow)
            case _:
                return NotImplemented

    def __add__(self, other: Any) -> DerivedColumn:
        match other:
            case Column():
                return DerivedColumn(self, other, op.add)
            case _:
                return NotImplemented

    def __sub__(self, other: Any) -> DerivedColumn:
        match other:
            case Column():
                return DerivedColumn(self, other, op.sub)
            case _:
                return NotImplemented

    def log10(self, unit_container: u.LogUnit = u.DexUnit) -> DerivedColumn:
        """
        Create a derived column that will compute the log of a given column. If
        the column contains units, the units must not be an astropy LogUnit
        (such as Dex or Mag)

        If you want the units of the new column to be a particular type of LogUnit,
        you can pass that type to the :code:`unit_container` argument. Defaults
        to DexUnit.
        """
        op = partial(_log10, unit_container=unit_container)
        return DerivedColumn(self, None, op)

    def exp10(self, expected_unit_container: u.LogUnit = u.DexUnit) -> DerivedColumn:
        """
        Create a derived column that will contain the base-10 exponentiation of the
        given column. If the column being exponentiated contains units, it must be an
        astropy LogUnit (e.g. Dex or Mag)

        You can specify the type of LogUnit container you expect the column to have with
        expected_unit_container. Defaults to DexUnit.
        """
        op = partial(_exp10, expected_unit_container=expected_unit_container)
        return DerivedColumn(self, None, op)

    def sqrt(self) -> DerivedColumn:
        """
        Create a derived column that will contain the square root of the given column.
        """
        return DerivedColumn(self, None, _sqrt)


class ConstructedColumn(Protocol):
    pass

    @property
    def requires(self) -> set[str]: ...
    @property
    def produces(self) -> Optional[set[str]]: ...
    @property
    def description(self) -> Optional[str]: ...

    def evaluate(
        self,
        data: dict[str, np.ndarray],
        index: DataIndex,
    ) -> np.ndarray | dict[str, np.ndarray]: ...

    def get_units(self, values: dict[str, u.Quantity]) -> dict[str, u.Unit]: ...


class DerivedColumn:
    """
    A derived column represents a combination of multiple columns that already exist in
    the dataset through multiplication or division by other columns or scalars, which
    may or may not have units of their own.

    In general this is dangerous, because we cannot necessarily infer how a particular
    unit is supposed to respond to unit transformations. For the moment, we only allow
    for combinations of columns that already exist in the dataset.

    In general, columns that exist in the dataset are materialized first. Derived
    columns are then computed from these. The order of creation of the derived columns
    must be kept constant, in case you get another column which is derived from a
    derived column.
    """

    def __init__(
        self,
        lhs: ColumnOrScalar,
        rhs: Optional[ColumnOrScalar],
        operation: Callable,
        description: Optional[str] = None,
    ):
        self.lhs = lhs
        self.rhs = rhs
        self.operation = operation
        self.description = description if description is not None else "None"

    @cached_property
    def requires(self):
        """
        Return the raw data columns required to make this column
        """
        vals = set()
        match self.lhs:
            case Column():
                vals.add(self.lhs.column_name)
            case DerivedColumn():
                vals = vals | self.lhs.requires
        match self.rhs:
            case Column():
                vals.add(self.rhs.column_name)
            case DerivedColumn():
                vals = vals | self.rhs.requires

        return vals

    @property
    def produces(self):
        return None

    def check_parent_existance(self, names: set[str]):
        match self.rhs:
            case Column():
                rhs_valid = self.rhs.column_name in names
            case DerivedColumn():
                rhs_valid = self.rhs.check_parent_existance(names)
            case _:
                rhs_valid = True

        match self.lhs:
            case Column():
                lhs_valid = self.lhs.column_name in names
            case DerivedColumn():
                lhs_valid = self.lhs.check_parent_existance(names)
            case _:
                lhs_valid = True

        return lhs_valid and rhs_valid

    def get_units(self, units: dict[str, u.Unit]):
        match self.lhs:
            case Column():
                lhs_unit = units[self.lhs.column_name]
            case DerivedColumn():
                lhs_unit = self.lhs.get_units(units)
            case _:
                lhs_unit = None
        match self.rhs:
            case Column():
                rhs_unit = units[self.rhs.column_name]
            case DerivedColumn():
                rhs_unit = self.rhs.get_units(units)
            case _:
                rhs_unit = None

        if self.operation in (op.sub, op.add) and (
            not isinstance(lhs_unit, u.LogUnit) or not isinstance(rhs_unit, u.LogUnit)
        ):
            if lhs_unit != rhs_unit:
                raise UnitsError("Cannot add/subtract columns with different units!")
            return lhs_unit

        match (lhs_unit, rhs_unit):
            case (None, None):
                return None
            case (_, None):
                if self.operation == op.pow:
                    return self.operation(lhs_unit, self.rhs)
                else:
                    return self.operation(lhs_unit, 1)
            case (None, _):
                return self.operation(1, rhs_unit)
            case (_, _):
                return self.operation(lhs_unit, rhs_unit)

    def combine_on_left(self, other: Column | DerivedColumn, operation: Callable):
        """
        Combine such that this column becomes the lhs of a new derived column.
        """
        match other:
            case Column() | DerivedColumn() | int() | float():
                return DerivedColumn(self, other, operation)
            case _:
                return NotImplemented

    def combine_on_right(self, other: Column | DerivedColumn, operation: Callable):
        """
        Combine such that this column becomes the rhs of a new derived column.
        """
        match other:
            case Column() | DerivedColumn() | int() | float():
                return DerivedColumn(other, self, operation)
            case _:
                return NotImplemented

    __mul__ = partialmethod(combine_on_left, operation=op.mul)
    __rmul__ = partialmethod(combine_on_right, operation=op.mul)
    __truediv__ = partialmethod(combine_on_left, operation=op.truediv)
    __rtruediv__ = partialmethod(combine_on_right, operation=op.truediv)
    __pow__ = partialmethod(combine_on_left, operation=op.pow)
    __add__ = partialmethod(combine_on_left, operation=op.add)
    __radd__ = partialmethod(combine_on_right, operation=op.add)
    __sub__ = partialmethod(combine_on_left, operation=op.sub)
    __rsub__ = partialmethod(combine_on_right, operation=op.sub)

    def log10(self, unit_container=u.DexUnit):
        op = partial(_log10, unit_container=unit_container)
        return DerivedColumn(self, None, op)

    def exp10(self, expected_unit_container: u.LogUnit = u.DexUnit):
        op = partial(_exp10, expected_unit_container=expected_unit_container)
        return DerivedColumn(self, None, op)

    def sqrt(self):
        return DerivedColumn(self, None, _sqrt)

    def evaluate(self, data: dict[str, np.ndarray], index: DataIndex) -> np.ndarray:
        lhs: np.typing.ArrayLike
        rhs: Optional[np.typing.ArrayLike]
        match self.lhs:
            case DerivedColumn():
                lhs = self.lhs.evaluate(data, index)
            case Column():
                lhs = data[self.lhs.column_name]
            case _:
                lhs = self.lhs
        match self.rhs:
            case DerivedColumn():
                rhs = self.rhs.evaluate(data, index)
            case Column():
                rhs = data[self.rhs.column_name]
            case _:
                rhs = self.rhs

        result = self.operation(lhs, rhs)
        return result


class EvaluatedColumn:
    def __init__(
        self,
        func: Callable,
        requires: set[str],
        produces: set[str],
        format: str,
        strategy: EvaluateStrategy = EvaluateStrategy.ROW_WISE,
        description: Optional[str] = None,
        **kwargs: Any,
    ):
        self.__func = func
        self.__requires = requires
        self.__kwargs = kwargs
        self.__produces = produces
        self.__format = format
        self.__strategy = strategy
        self.description = description

    @property
    def requires(self):
        return copy(self.__requires)

    @property
    def produces(self):
        return copy(self.__produces)

    @property
    def signature(self):
        return signature(self.__func)

    @property
    def kwarg_names(self):
        return self.__kwargs.keys()

    def get_units(self, units: dict[str, np.ndarray]):
        test_data: dict[str, Any]
        match self.__strategy:
            case EvaluateStrategy.ROW_WISE:
                test_data = {
                    name: np.random.randint(20, 40) for name in self.__requires
                }
            case _:
                test_data = {
                    name: np.random.randint(20, 40, 2) for name in self.__requires
                }

        if self.__format == "astropy":
            test_data = {
                name: td * units[name] if units.get(name) is not None else td
                for name, td in test_data.items()
            }

        results = self.__func(**test_data, **self.__kwargs)
        if not isinstance(results, dict):
            results = {self.__func.__name__: results}

        return {
            name: result.unit if isinstance(result, u.Quantity) else None
            for name, result in results.items()
        }

    def evaluate(self, data: dict[str, np.ndarray], index: DataIndex):
        data = {name: data[name] for name in self.__requires}
        if self.__format != "astropy":
            data = {
                name: val.value if isinstance(val, u.Quantity) else val
                for name, val in data.items()
            }
        match self.__strategy:
            case EvaluateStrategy.VECTORIZE:
                return evaluate_vectorized(data, self.__func, self.__kwargs)
            case EvaluateStrategy.ROW_WISE:
                return evaluate_rows(data, self.__func, self.__kwargs)
            case EvaluateStrategy.CHUNKED:
                if not isinstance(index, tuple):
                    raise ValueError(
                        "Cannot evaluate in CHUNKED strategy with a non-chunked index"
                    )
                return evaluate_chunks(data, self.__func, self.__kwargs, index)

    def evaluate_one(self, dataset: Dataset):
        match self.__strategy:
            case EvaluateStrategy.VECTORIZE:
                values = (
                    dataset.select(self.__requires)
                    .take(1)
                    .get_data(self.__format, unpack=False)
                )
                values = dict(values)
                return self.__func(**values, **self.__kwargs)

            case EvaluateStrategy.ROW_WISE:
                values = (
                    dataset.select(self.__requires)
                    .take(1)
                    .get_data(self.__format, unpack=True)
                )
                values = dict(values)
                return self.__func(**values, **self.__kwargs)

            case EvaluateStrategy.CHUNKED:
                index = dataset.index
                assert isinstance(index, tuple)
                first_chunk_size = index[1][0]
                first_chunk = (
                    dataset.select(self.__requires)
                    .take(first_chunk_size)
                    .get_data(self.__format)
                )
                first_chunk = dict(first_chunk)
                return self.__func(**first_chunk, **self.__kwargs)

        pass


class ColumnMask:
    """
    A mask is a class that represents a mask on a column. ColumnMasks evaluate
    to t/f for every element in the given column.
    """

    def __init__(
        self,
        column_name: str,
        value: float | u.Quantity,
        operator: Callable[[table.Column, float | u.Quantity], np.ndarray],
    ):
        self.column_name = column_name
        self.value = value
        self.operator = operator

    @property
    def requires(self):
        return {self.column_name}

    def apply(self, column: u.Quantity | np.ndarray) -> np.ndarray:
        """
        mask the dataset based on the mask.
        """
        # Astropy's errors are good enough here
        if isinstance(column, table.Table):
            column = column[self.column_name]

        if isinstance(self.value, u.Quantity) and isinstance(column, u.Quantity):
            if self.value.unit != column.unit:
                raise ValueError(
                    f"Incompatible units in fiter: {self.value.unit} and {column.unit}"
                )

        elif isinstance(column, u.Quantity):
            return self.operator(column.value, self.value)

        return self.operator(column, self.value)  # type: ignore

    def __and__(self, other: Self | CompoundColumnMask):
        return CompoundColumnMask(self, other, lambda left, right: left & right)

    def __or__(self, other: Self | CompoundColumnMask):
        return CompoundColumnMask(self, other, lambda left, right: left | right)


class CompoundColumnMask:
    def __init__(
        self,
        left: ColumnMask | Self,
        right: ColumnMask | Self,
        op: Callable[[np.ndarray, np.ndarray], np.ndarray],
    ):
        self.__left = left
        self.__right = right
        self.__op = op

    @property
    def requires(self):
        columns = set()
        columns |= self.__left.requires
        columns |= self.__right.requires
        return columns

    def __and__(self, other: ColumnMask | Self):
        return CompoundColumnMask(self, other, lambda left, right: left & right)

    def __or__(self, other: ColumnMask | Self):
        return CompoundColumnMask(self, other, lambda left, right: left | right)

    def apply(self, data):
        left_mask = self.__left.apply(data)
        right_mask = self.__right.apply(data)
        return self.__op(left_mask, right_mask)



from __future__ import annotations

from enum import Enum
from functools import reduce
from typing import TYPE_CHECKING, Any, Callable, Optional, Protocol

import numpy as np

from opencosmo.index import get_data, get_length

if TYPE_CHECKING:
    import h5py
    from numpy.typing import DTypeLike

    from opencosmo.index import DataIndex
    from opencosmo.mpi import MPI


class ColumnCombineStrategy(Enum):
    SUM = "sum"
    CONCAT = "concat"


class ColumnSource(Protocol):
    def __len__(self) -> int: ...
    @property
    def shape(self) -> tuple[int, ...]: ...
    @property
    def dtype(self) -> DTypeLike: ...
    @property
    def data(self) -> np.ndarray: ...


class ColumnWriter:
    def __init__(
        self,
        column_sources: list[ColumnSource],
        combine_strategy: ColumnCombineStrategy,
        attrs: dict[str, Any] = {},
    ):
        self.__sources = column_sources
        self.__combine_strategy = combine_strategy
        self.__attrs = attrs
        dtypes = set(map(lambda source: source.dtype, self.__sources))
        if len(dtypes) > 1:
            raise ValueError("A single column can not have multiple data types!")
        self.__dtype = dtypes.pop()
        self.__transformation: Optional[Callable] = None

    @classmethod
    def from_numpy_array(
        cls,
        data: np.ndarray,
        strategy: ColumnCombineStrategy = ColumnCombineStrategy.CONCAT,
        attrs: dict[str, Any] = {},
    ):
        source = NumpySource(data)
        return ColumnWriter([source], strategy, attrs)

    @classmethod
    def from_h5_dataset(
        cls,
        dataset: h5py.Dataset,
        index: DataIndex,
        strategy: ColumnCombineStrategy = ColumnCombineStrategy.CONCAT,
        attrs: dict[str, Any] = {},
    ):
        source = Hdf5Source(dataset, index)
        return ColumnWriter([source], strategy, attrs)

    def combine(self, others: list[ColumnWriter]):
        new_sources = reduce(
            lambda acc, other: acc + other.__sources, others, self.__sources
        )
        new_sources = list(filter(lambda s: len(s) > 0, new_sources))
        return ColumnWriter(new_sources, self.combine_strategy, self.__attrs)

    def set_transformation(
        self, transformation: Callable[[np.ndarray, Optional[MPI.Comm]], np.ndarray]
    ):
        if self.__transformation is not None:
            raise ValueError(
                "A transformation can only be set on a column writer a single time!"
            )
        self.__transformation = transformation

    def __len__(self):
        match self.combine_strategy:
            case ColumnCombineStrategy.CONCAT:
                return sum(len(source) for source in self.__sources)
            case ColumnCombineStrategy.SUM:
                return len(self.__sources[0])

    @property
    def has_transformation(self):
        return self.__transformation is not None

    @property
    def shape(self):
        shape = self.__sources[0].shape
        return (len(self),) + shape[1:]

    @property
    def combine_strategy(self) -> ColumnCombineStrategy:
        return self.__combine_strategy

    @property
    def dtype(self) -> DTypeLike:
        return self.__dtype

    @property
    def attrs(self) -> dict[str, Any]:
        return self.__attrs

    @property
    def data(self) -> np.ndarray:
        return self.get_data()

    def get_data(self, comm: Optional[MPI.Comm] = None):
        match self.combine_strategy:
            case ColumnCombineStrategy.CONCAT:
                data = np.concatenate([source.data for source in self.__sources])
            case ColumnCombineStrategy.SUM:
                data = np.vstack([source.data for source in self.__sources]).sum(axis=0)
        if self.__transformation is not None:
            data = self.__transformation(data, comm=comm)  # type: ignore
        return data


class Hdf5Source:
    def __init__(self, h5ds: h5py.Dataset, index: DataIndex):
        self.__source = h5ds
        self.__index = index

    def __len__(self):
        return get_length(self.__index)

    @property
    def shape(self):
        return (len(self),) + self.__source.shape[1:]

    @property
    def dtype(self):
        return self.__source.dtype

    @property
    def data(self):
        data = get_data(self.__source, self.__index)
        return data


class NumpySource:
    def __init__(self, arr: np.ndarray):
        self.__source = arr

    def __len__(self):
        return len(self.__source)

    @property
    def shape(self):
        return self.__source.shape

    @property
    def dtype(self):
        return self.__source.dtype

    @property
    def data(self):
        return self.__source



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from . import into_array

if TYPE_CHECKING:
    from opencosmo.index import ChunkedIndex, DataIndex, SimpleIndex


def project(source: DataIndex, other: DataIndex):
    match (source, other):
        case (tuple(), np.ndarray()):
            return __project_simple_on_chunked(source, other)
        case (tuple(), tuple()):
            return __project_chunked_on_chunked(source, other)
        case (np.ndarray(), np.ndarray()):
            return __project_simple_on_simple(source, other)
        case (np.ndarray(), tuple()):
            return __project_chunked_on_simple(source, other)


def __project_simple_on_simple(source, other: DataIndex):
    isin = np.isin(source, other)
    return np.where(isin)[0]


def __project_chunked_on_simple(source, other: DataIndex):
    return project(source, into_array(other))


def __project_simple_on_chunked(source: ChunkedIndex, other: SimpleIndex):
    return project(into_array(source), other)


def __project_chunked_on_chunked(source: ChunkedIndex, other: ChunkedIndex):
    source_ends = source[0] + source[1]
    other_ends = other[0] + other[1]

    clipped_starts = np.clip(other[0][:, np.newaxis], a_min=source[0], a_max=None)
    clipped_ends = np.clip(other_ends[:, np.newaxis], a_min=None, a_max=source_ends)

    rs = np.cumsum(source[1])
    rs = np.insert(rs, 0, 0)[:-1]

    starts = clipped_starts - source[0] + rs
    ends = clipped_ends - source[0] + rs
    sizes = ends - starts

    starts = starts.flatten()
    sizes = sizes.flatten()

    mask = sizes > 0
    return (starts[mask], sizes[mask])



from __future__ import annotations

from importlib import import_module

import astropy.units as u
import numpy as np
from astropy.table import Column, QTable


def verify_format(output_format: str):
    match output_format:
        case "astropy":
            return
        case "numpy":  # these two are core dependencies
            return
        case "pandas":
            import_name = "pandas"
        case "arrow":
            import_name = "pyarrow"
        case "polars":
            import_name = "polars"
        case _:
            raise ValueError(f"Unknown data output format {output_format}")

    __verify_import(import_name, output_format)


def __verify_import(import_name: str, format_name: str):
    try:
        import_module(import_name)
    except ImportError as e:
        raise ImportError(
            f"Data was requested in {format_name} format but could not import {import_name} package. Got '{e}'"
        )


def convert_data(data: dict[str, np.ndarray], output_format: str):
    match output_format:
        case "astropy":
            return __convert_to_astropy(data)
        case "numpy":
            return __convert_to_numpy(data)
        case "pandas":
            return __convert_to_pandas(data)
        case "polars":
            return __convert_to_polars(data)
        case "arrow":
            return __convert_to_arrow(data)
        case _:
            raise ValueError(f"Unknown data output format {output_format}")


def __convert_to_astropy(data: dict[str, np.ndarray]) -> QTable:
    if len(data) == 1:
        return next(iter(data.values()))
    if any(
        (isinstance(d, u.Quantity) and d.isscalar) or not isinstance(d, np.ndarray)
        for d in data.values()
    ):
        return data

    return QTable(data, copy=False)


def __convert_to_numpy(
    data: dict[str, np.ndarray],
) -> dict[str, np.ndarray] | np.ndarray:
    converted_data = dict(
        map(
            lambda kv: (
                kv[0],
                kv[1].value if isinstance(kv[1], (u.Quantity, Column)) else kv[1],
            ),
            data.items(),
        )
    )
    if len(converted_data) == 1:
        return next(iter(converted_data.values()))
    return converted_data


def __convert_to_pandas(data: dict[str, np.ndarray]):
    import pandas as pd

    numpy_data = __convert_to_numpy(data)
    if isinstance(numpy_data, np.ndarray):  # only one column
        return pd.Series(numpy_data, name=next(iter(data.keys())))

    return pd.DataFrame(numpy_data, copy=True)


def __convert_to_arrow(data: dict[str, np.ndarray]):
    import pyarrow as pa  # type: ignore

    numpy_data = __convert_to_numpy(data)
    if isinstance(numpy_data, np.ndarray):
        return pa.array(numpy_data)

    converted_data = map(
        lambda kv: (kv[0], pa.array(kv[1])),
        data.items(),
    )
    return dict(converted_data)


def __convert_to_polars(data: dict[str, np.ndarray]):
    import polars as pl

    numpy_data = __convert_to_numpy(data)
    if isinstance(numpy_data, np.ndarray):
        return pl.Series(name=next(iter(data.keys())), values=numpy_data)

    return pl.from_dict(data)  # type: ignore



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Iterable, Optional

import healpy as hp
import numpy as np

from opencosmo import dataset as ds
from opencosmo.io.mpi import get_all_keys
from opencosmo.io.schema import FileEntry, make_schema
from opencosmo.mpi import get_comm_world
from opencosmo.spatial.check import find_coordinates_2d

if TYPE_CHECKING:
    from opencosmo.io.schema import Schema
    from opencosmo.mpi import MPI


def update_order(data: np.ndarray, comm: Optional[MPI.Comm], order: np.ndarray):
    if comm is not None:
        return update_global_order_mpi(data, comm, order)

    return data[order]


def update_global_order_mpi(data, comm, order):
    needs_global_reordering = comm.allgather(np.any((order < 0) | (order > len(order))))
    if not np.any(needs_global_reordering):
        return data[order]

    ends = np.cumsum(comm.allgather(len(order)))
    starts = np.insert(ends, 0, 0)
    global_order = order + starts[comm.Get_rank()]
    all_data = comm.allgather(data)
    return np.concat(all_data)[global_order]


def sync_metadata(dataset_schemas: list[Schema]):
    additional_metadata = [schema.attributes for schema in dataset_schemas]
    if not any(additional_metadata):
        return {}
    if not all(am == additional_metadata[0] for am in additional_metadata[1:]):
        raise ValueError("Datasets don't have the same metadata!")

    return additional_metadata[0]


def sync_headers(datasets: list[ds.Dataset], redshift_range):
    if not datasets and (comm := get_comm_world()) is not None:
        comm.allgather(0)
        comm.allgather(-1)
        comm.allgather((1000, -1))
        return

    steps = (
        dataset.header.file.step
        for dataset in datasets
        if dataset.header.file.step is not None
    )
    redshifts = (
        dataset.header.file.redshift
        for dataset in datasets
        if dataset.header.file.redshift is not None
    )
    step = max(steps)
    redshift = max(redshifts)

    if (comm := get_comm_world()) is not None:
        step = np.max(comm.allgather(step))
        redshift = max(comm.allgather(redshift))
        z_ranges = comm.allgather(redshift_range)
        z_min = min(zr[0] for zr in z_ranges)
        z_max = max(zr[1] for zr in z_ranges)
        redshift_range = (z_min, z_max)

    # lightcones are identified by their upper redshift slice
    header_schema = datasets[0].header.dump()
    header_schema.attributes["file"]["redshift"] = redshift
    header_schema.attributes["file"]["step"] = step
    header_schema.attributes["lightcone"]["z_range"] = redshift_range
    return header_schema


def stack_lightcone_datasets_in_schema(
    datasets: dict[str, list[ds.Dataset]],
    name: Optional[str],
    redshift_range: Optional[tuple[float, float]],
):
    n_datasets = sum(len(lst) for lst in datasets.values())
    if n_datasets == 1 and get_comm_world() is None:
        dataset_list = next(iter(datasets.values()))

        schema = dataset_list[0].make_schema(name=name)
        header = sync_headers(dataset_list, redshift_range)
        schema.children["header"] = header
        return {"data": schema}

    schema_children = {}
    ds_groups = get_all_keys(datasets, get_comm_world())
    for ds_group in ds_groups:
        ds_list = datasets.get(ds_group, [])
        ds_list = list(filter(lambda ds: len(ds) > 0, ds_list))
        if len(ds_list) == 0:
            get_stacked_lightcone_order([], -1)
            sync_headers(ds_list, None)
            continue
        schemas = [ds.make_schema(name=name) for ds in ds_list]
        index_names = list(schemas[0].children["index"].children.keys())
        index_names.sort()
        max_level = int(index_names[-1][-1])

        assert all(isinstance(dataset, ds.Dataset) for dataset in ds_list)
        new_data_group = stack_data_groups(
            [schema.children["data"] for schema in schemas]
        )

        order = get_stacked_lightcone_order(ds_list, max_level)
        updater = partial(update_order, order=order)

        for column in new_data_group.columns.values():
            column.set_transformation(updater)

        new_index_group = stack_index_groups(
            [schema.children["index"] for schema in schemas]
        )
        header_schema = sync_headers(ds_list, redshift_range)
        additional_metadata = sync_metadata(schemas)

        children = {
            "data": new_data_group,
            "index": new_index_group,
            "header": header_schema,
        }

        schema_name = ds_group if len(datasets) > 1 else name
        assert schema_name is not None
        schema_children[schema_name] = make_schema(
            schema_name,
            FileEntry.LIGHTCONE,
            children=children,
            attributes=additional_metadata,
        )

    return schema_children


def stack_index_groups(schemas: list[Schema]):
    base_schema = schemas[0]
    new_children = {}
    for index_level in base_schema.children.keys():
        all_level_schemas = [schema.children[index_level] for schema in schemas]
        new_children[index_level] = stack_data_groups(all_level_schemas)
    return make_schema("index", FileEntry.COLUMNS, new_children)


def stack_data_groups(schemas: list[Schema]):
    if len(schemas) == 1:
        return schemas[0]
    base_schema = schemas[0]
    new_writers = {}
    for name, column_writer in base_schema.columns.items():
        other_writers = [schema.columns[name] for schema in schemas[1:]]
        new_writer = column_writer.combine(other_writers)
        new_writers[name] = new_writer

    new_schema = make_schema(
        base_schema.name,
        base_schema.type,
        children={},
        columns=new_writers,
        attributes=base_schema.attributes,
    )
    return new_schema


def get_order_mpi(pixels, comm):
    pixel_order = np.argsort(pixels)
    if len(pixels) > 0:
        pixel_ranges = comm.allgather((pixels[pixel_order[0]], pixels[pixel_order[-1]]))

    else:
        pixel_ranges = comm.allgather(None)

    pixel_ranges = [pr for pr in pixel_ranges if pr is not None]
    for i in range(len(pixel_ranges) - 1):
        if pixel_ranges[i][1] > pixel_ranges[i + 1][0]:
            break
    else:
        return pixel_order

    all_pixels = np.concat(comm.allgather(pixels))
    new_order = np.argsort(all_pixels)
    bounds = np.cumsum(comm.allgather(len(pixels)))
    bounds = np.insert(bounds, 0, 0)
    rank = comm.Get_rank()
    return new_order[bounds[rank] : bounds[rank + 1]] - bounds[rank]


def get_stacked_lightcone_order(datasets: Iterable[ds.Dataset], max_index_depth: int):
    datasets = list(datasets)
    nside = 2**max_index_depth
    coordinates = list(map(find_coordinates_2d, datasets))
    coordinates = list(filter(lambda coord_list: len(coord_list) > 0, coordinates))

    if datasets:
        pixels = np.concatenate(
            [
                hp.ang2pix(
                    nside, coords.ra.value, coords.dec.value, lonlat=True, nest=True
                )
                for coords in coordinates
            ]
        )

    else:
        pixels = np.array([])

    if (comm := get_comm_world()) is not None:
        return get_order_mpi(pixels, comm)

    return np.argsort(pixels)



from __future__ import annotations

from functools import reduce, wraps
from typing import TYPE_CHECKING, Callable, Iterable

import astropy.units as u

if TYPE_CHECKING:
    from .column import Column, DerivedColumn

from .column import col


def into_cols(func: Callable):
    @wraps(func)
    def wrapper(*columns: str | Column | DerivedColumn, **kwargs):
        new_columns = tuple(
            map(
                lambda colname: col(colname) if isinstance(colname, str) else colname,
                columns,
            )
        )
        return func(*new_columns, **kwargs)

    return wrapper


def offset_3d(
    coord_name_a: str, coord_name_b: str, labels: Iterable[str] = ["x", "y", "z"]
):
    """
    Create a derived column that contains the magnitude of the offset between two sets of 3d coordinates.
    For exmaple, to get the magnitude of the difference between the FoF halo centers and the SOD halo centers:

    .. code-block:: python

        from opencosmo.column import offset_3d
        import opencosmo as oc

        dataset = oc.open("haloproperties.hdf5")

        offset_column = offset_3d("fof_halo_com", "sod_halo_com")
        dataset = dataset.with_new_columns(offset=offset_column)


    This function assumes that the columns are named "fof_halo_center_{x, y, z}" and "sod_halo_center_{x, y, z}",
    you can choose different labels by setting the :code:`labels` argument.

    This function outputs a derived column that can be passed into :py:meth:`with_new_columns <opencosmo.Dataset.with_new_files>. This function will never fail, but :py:meth:`with_new_columns <opencosmo.Dataset.with_new_columns>`
    will if the columns do not all have the same units.

    Parameters
    ----------
    coord_name_a: str
        The base name of the first coordinate

    coord_name_b: str
        The base name of the second coordinate

    labels: Iterable[str], default = ["x", "y", "z"]
        The coordinate labels. The names of the columns are assumed to be "{coord_name_a}_{labels}" and
        "{coord_name_b}_{labels}"

    """
    delta_coords = tuple(
        map(
            lambda label: col(f"{coord_name_a}_{label}")
            - col(f"{coord_name_b}_{label}"),
            labels,
        )
    )
    return norm_cols(*delta_coords)


@into_cols
def add_mag_cols(*magnitudes: Column | DerivedColumn):
    """
    Add together any number of magnitude columns to get a total magnitude. This function
    takes in the names of the magnitude columns, and produces a DerivedColumn that can be
    passed into :py:meth:`with_new_columns <opencosmo.Dataset.with_new_columns>`

    This function will never fail, but :py:meth:`with_new_columns <opencosmo.Dataset.with_new_columns>`
    will if you include columns that are not magnitudes.

    .. code-block:: python

        import opencosmo as oc
        from opencosmo.column import add_mag_cols

        dataset = oc.open("catalog.hdf5")
        mag_total = add_mag_cols("mag_g", "mag_r", "mag_i", "mag_z", "mag_y")

        dataset = dataset.with_new_columns(mag_total=mag_total)

    Parameters
    ----------
    *magnitudes: str | Column | DerivedColumn
        Any number of magnitude columns. You can pass in simple column names, columns constructred
        with :py:meth:`opencosmo.col`, or columns created from combinations of other columns

    Returns
    -------
    new_column: DerivedColumn
        A new derived column that can be passed into :py:meth:`with_new_columns <opencosmo.Dataset.with_new_columns>`
    """
    if len(magnitudes) < 2:
        raise ValueError("Expected at least two magnitudes to add together!")
    fluxes = map(
        lambda column: (-0.4 * column).exp10(expected_unit_container=u.MagUnit),
        magnitudes,
    )
    total_flux = next(fluxes)
    for flux in fluxes:
        total_flux += flux

    return -2.5 * total_flux.log10(u.MagUnit)


@into_cols
def norm_cols(*columns: Column | DerivedColumn):
    """
    Get the euclidian norm of any number of columns. This function takes in the names
    of the magnitude columns, and produces a DerivedColumn that can be passed into
    :py:meth:`with_new_columns <opencosmo.Dataset.with_new_columns>`

    This function will never fail, but :py:meth:`with_new_columns <opencosmo.Dataset.with_new_columns>`
    will if the columns do not have the same units.

    Parameters
    ----------
    *columns: str | Column | DerivedColumn
        Any number of columns. You can pass in simple column names, columns constructred
        with :py:meth:`opencosmo.col`, or columns created from combinations of other columns

    Returns
    -------
    new_column: DerivedColumn
        A new derived column that can be passed into :py:meth:`with_new_columns <opencosmo.Dataset.with_new_columns>`
    """
    if len(columns) < 2:
        raise ValueError("Expected at least two magnitudes to add together!")

    squared_columns = map(lambda column: column**2, columns)
    sum_squared = reduce(lambda acc, col_sq: acc + col_sq, squared_columns)
    return sum_squared.sqrt()



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, NamedTuple, Optional

if TYPE_CHECKING:
    from .writer import ColumnWriter


class FileEntry(Enum):
    DATASET = "dataset"
    MULTI_DATASET = "multi_dataset"
    STRUCTURE_COLLECTION = "structure_collection"
    SIMULATION_COLLECTION = "simulation_collection"
    LIGHTCONE = "lightcone"
    LIGHCONE_MAP = "lightcone_map"
    HEALPIX_MAP = "healpix_map"
    COLUMNS = "columns"
    METADATA = "metadata"
    EMPTY = "empty"


class Schema(NamedTuple):
    name: str
    type: FileEntry
    children: dict[str, Schema]
    columns: dict[str, ColumnWriter]
    attributes: dict[str, Any]


def empty_schema(name: str, type_: FileEntry) -> Schema:
    return Schema(name, type_, {}, {}, {})


def make_schema(
    name: str,
    type_: FileEntry,
    children: Optional[dict] = None,
    columns: Optional[dict] = None,
    attributes: Optional[dict] = None,
):
    if children is None:
        children = {}
    if columns is None:
        columns = {}
    if attributes is None:
        attributes = {}
    return Schema(name, type_, children, columns, attributes)
